#!/usr/bin/python3
#
# Univention Management Console
#  Administration of Two Factor (2FA) Auth
#
# Copyright 2020-2022 Univention GmbH
#
# https://www.univention.de/
#
# All rights reserved.
#
# The source code of this program is made available
# under the terms of the GNU Affero General Public License version 3
# (GNU AGPL V3) as published by the Free Software Foundation.
#
# Binary versions of this program provided by Univention to you as
# well as other copyrighted, protected or trademarked materials like
# Logos, graphics, fonts, specific documentations and configurations,
# cryptographic keys etc. are subject to a license agreement between
# you and Univention and not subject to the GNU AGPL V3.
#
# In the case you use this program under the terms of the GNU AGPL V3,
# the program is provided in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License with the Debian GNU/Linux or Univention distribution in file
# /usr/share/common-licenses/AGPL-3; if not, see
# <https://www.gnu.org/licenses/>.
import json

import univention.config_registry as ucr
from ldap.filter import escape_filter_chars, filter_format
from six.moves import configparser
from ucsschool.lib.models.user import School, User
from ucsschool.lib.school_umc_base import SchoolBaseModule
from ucsschool.lib.school_umc_ldap_connection import LDAP_Connection
from univention.lib.i18n import Translation
from univention.lib.misc import custom_username, custom_groupname
from univention.management.console.error import UMC_Error
from univention.management.console.log import MODULE
from univention.management.console.modules.decorators import sanitize
from univention.management.console.modules.decorators import simple_response
from univention.management.console.modules.sanitizers import StringSanitizer, IntegerSanitizer
from univention.udm import UDM, UdmError

from .exceptions import TwofactorException, UnAuthorizedException, NoSchoolSelectedException, NoSchoolAvailableException, ModuleTimeoutException
from .twofactor_admin import get_auth_token, disable_or_delete_all_tokens_for_user, rollout_qr_token, rollover_hardware_token, lost_token, return_hardware_token

_ = Translation('dataport-schuposh-twofactor-admin_module').translate


class Instance(SchoolBaseModule):

    def __init__(self):
        super(Instance, self).__init__()
        self.url = ""
        self.auth_token = ""
        self.current_school = None
        # all the schools the current user is admin of (gets to do token actions for)
        self.admin_schools = []
        self.is_global_admin = False

    def is_user_in_group(self, groupname):
        try:
            group = UDM.admin().version(1).get('groups/group').get_by_id(groupname)
            in_group = self.user_dn.lower() in [dn.lower() for dn in group.props.users]
            return in_group
        except UdmError as e:
            MODULE.error("UDM error during user group check: %s" % e.msg)
            return False

    def init(self):
        try:
            SchoolBaseModule.init(self)
            self.config = ucr.ConfigRegistry()
            self.config.load()
            self.url = self.config.get("privacyidea/saml/url")
            if self.url is None:
                raise UMC_Error("Can't find privacyidea/saml/url UCR variable")
            """
            The basic Authorization is done by ucs internals, a user needs the 2FA admin (twofactor-admin-all) operation set in one of his groups
            (nested groups are not supported)
            Since allmighty / domain Admins can see non-ucs@school users (which school Admins can't), we need to query the groups
            again to decide which user list to present
            """
            self.is_global_admin = (
                self.username == custom_username("Administrator") or
                self.is_user_in_group(custom_groupname('Domain Admins')) or
                self.is_user_in_group(custom_groupname('iqsh-admins')) or
                self.is_user_in_group(custom_groupname('dataport-schuposh-fvm-admins'))
            )

            # Create a list of schools which the user is able to administrate
            # this does not apply to global admins
            if not self.is_global_admin:
                user_obj = UDM(self.get_user_ldap_connection()).version(1).get('users/user').get(self.user_dn)
                for ou in user_obj.props.school:
                    admin_dn = "cn=admins-%s,cn=ouadmins,cn=groups,%s" % (ou.lower(), self.config.get("ldap/base"))
                    if admin_dn in user_obj.props.groups:
                        self.admin_schools.append(ou.lower())

            # get privacyIDEA Credentials from /etc/twofactor-admin.secret
            config = configparser.RawConfigParser()
            config.read('/etc/twofactor-admin.secret')

            self.auth_token = get_auth_token(
                self.url,
                config.get('privacyIDEA', "user"),
                config.get('privacyIDEA', "pass")
            )
        except TwofactorException as e:
            MODULE.error("Init failed for logged in user %s: %s" % (self.username, e))
            raise UMC_Error(e.message)
        except configparser.NoSectionError:
            MODULE.error("PricvacyIDEA Credentials not found")
            raise UMC_Error(_("PrivacyIDEA Credential Configuration Error, please contact Administration."))

    @simple_response
    def get_is_admin(self):
        return {"is_admin": self.is_global_admin}

    def check_permissions(self, target_uid):
        """Return None means allow, otherwise exception is raised."""
        # Admin is allmighty
        if self.is_global_admin:
            return

        # This seems weird, but there's no hook for processing timeouts.
        # So our only way to determine a timeout is to realise that
        # there is no selected school for an action which requires a selected school.
        if not self.current_school:
            MODULE.error("No current school set, possible Timeout")
            raise ModuleTimeoutException()

        if self.current_school not in self.admin_schools:
            MODULE.error("User %s is not authorized to administrate school %s" % (self.username, self.current_school))
            raise UnAuthorizedException()

        user_objs = list(UDM(self.get_user_ldap_connection()).version(1).get('users/user').search(filter_format("(&(uid=%s)(ucsschoolSchool=%s))", (target_uid, self.current_school))))
        if len(user_objs) > 1:
            raise UMC_Error("Found multiple results for uid %s" % target_uid)

        if len(user_objs) == 0:
            raise UMC_Error("Found no results for uid %s" % target_uid)

        user_obj = user_objs[0]

        if any(school.lower() in self.admin_schools for school in user_obj.props.school):
            return
        MODULE.error("Target User %s is not member of current school (%s)" % (target_uid, self.current_school))
        raise UnAuthorizedException

    @sanitize(
        uid=StringSanitizer()
    )
    @simple_response
    def rollover_qr_token(self, uid):
        try:
            self.check_permissions(uid)
            disable_or_delete_all_tokens_for_user(uid, self.url, self.auth_token)
            otp = rollout_qr_token(self.url, self.auth_token, uid)
            MODULE.info("QR Token for User %s generated by %s" % (uid, self.user_dn))
            return otp
        except TwofactorException as e:
            MODULE.error("Error rolling over qr Token, logged in user %s, target user %s:  %s" % (self.username, uid, e.message))
            raise UMC_Error(e.message)

    @sanitize(
        pattern=StringSanitizer(required=False)
    )
    @simple_response
    def get_users(self, pattern=None):
        try:
            users = []
            # FIXME: put this filter to an UCRV?
            pattern_filter = "(&(uid=*%s*)(|(objectClass=ucsschoolAdministrator)(objectClass=ucsschoolTeacher)))" % (
                escape_filter_chars(pattern),) if pattern else ""

            if self.is_global_admin:
                user_mod = UDM(self.get_user_ldap_connection()).version(1).get("users/user")
                for obj in user_mod.search(pattern_filter):
                    users.append({"uid": obj.props.username, "firstname": obj.props.firstname, "lastname": obj.props.lastname})

            else:
                if self.current_school is None:
                    raise NoSchoolSelectedException()

                lo = self.get_user_ldap_connection()
                # Found no easy way to do this
                filter_str = "(&%s(|(objectClass=ucsschoolAdministrator)(objectClass=ucsschoolTeacher))" \
                             "(ucsschoolSchool=%s))" % (
                                 pattern_filter, escape_filter_chars(self.current_school))

                udm_users = User.get_all(lo, self.current_school, filter_str=filter_str)
                users = [{"uid": udm_user.name, "firstname": udm_user.firstname, "lastname": udm_user.lastname}
                         for udm_user in udm_users]

            return users

        except TwofactorException as e:
            MODULE.error("Error fetching user list for logged in user %s:  %s" % (self.username, e.message))
            raise UMC_Error(e.message)

    @LDAP_Connection()
    def get_schools(self, request, ldap_user_read=None):
        try:
            # only returns school for this user...
            ldap_school = School.from_binddn(ldap_user_read)
            schools = [{"label": school.display_name, "id": school.name} for school in ldap_school]
            # filter out schools for which the user is no admin
            schools2 = filter(lambda x: x['id'].lower() in self.admin_schools, schools)
            l = list(schools2)
            if len(l) == 0:
                raise NoSchoolAvailableException()
            self.finished(request.id, l)
        except TwofactorException as e:
            MODULE.error("Error fetching schools for user %s:  %s" % (self.username, e.message))
            raise UMC_Error(e.message)

    @sanitize(
        school=StringSanitizer(required=True)
    )
    @simple_response
    def set_current_school(self, school):
        try:
            if school.lower() not in self.admin_schools:
                raise UnAuthorizedException()
            self.current_school = school.lower()
            MODULE.info("Set current school %s for user %s" % (self.current_school, self.user_dn))
            return "success"
        except TwofactorException as e:
            MODULE.error("Error setting current school (user %s): %s" % (self.username, e.message))
            raise UMC_Error(e.message)

    @sanitize(
        uid=StringSanitizer(required=True),
        serial=IntegerSanitizer(required=True),
        # OTPs can have leading zeros, if the type is not string, there's a random error...
        current_otp_value=StringSanitizer(required=True)
    )
    @simple_response
    def assign_hardware_token(self, uid, serial, current_otp_value):
        try:
            self.check_permissions(uid)
            rollover_hardware_token(self.url, self.auth_token, uid, serial, current_otp_value)
            MODULE.info("Hardware Token %s was assigned to User %s by %s" % (serial, uid, self.user_dn))
            return "success"
        except TwofactorException as e:
            MODULE.error("Error assigning Hardware Token %s for user %s: %s" % (serial, uid, e.message))
            raise UMC_Error(e.message)

    @sanitize(
        uid=StringSanitizer(required=True)
    )
    @simple_response
    def lost_token(self, uid):
        try:
            self.check_permissions(uid)
            lost_token(self.url, self.auth_token, uid)
            MODULE.info("Token for User %s was marked as lost by %s" % (uid, self.user_dn))
            return "success"
        except TwofactorException as e:
            MODULE.error("reporting lost token for user %s: %s" % (uid, e.message))
            raise UMC_Error(e.message)

    @sanitize(
        serial=IntegerSanitizer(required=True),
        # OTPs can have leading zeros, if the type is not string, there's a random error...
        current_otp_value=StringSanitizer(required=True)
    )
    @simple_response
    def return_hardware_token(self, serial, current_otp_value):
        try:
            return_hardware_token(self.url, self.auth_token, serial, current_otp_value)
            MODULE.info("Return of Hardware Token %s was facilitated by %s" % (serial, self.user_dn))
            return "success"
        except TwofactorException as e:
            MODULE.error("Error returning hardware Token %s: %s" % (serial, e.message))
            raise UMC_Error(e.message)
