# This has no dependencies to UMC on purpose to make testing less complicated
import requests
from requests import ConnectionError

from .exceptions import *

def disable_or_delete_all_tokens_for_user(user, url, auth_token, disable=False, strict=False):
    try:
        # search all tokens for teacher
        r = requests.get(url + "/token", headers={'Authorization': auth_token}, json={'user': user})
        result = r.json()['result']
        # There are tokens for the user to delete
        if r.status_code == 200:
            if strict and result['value']['count'] == 0:
                raise UserHasNoActiveTokensException()
            #TODO: There seems to be some paging involved, for example only 15ish tokens are deleted in one go
            #Since we only allow one token per user this is not an issue atm, or at least not until somebody
            #looses 15 hardware token.
            # Hint: the parameter "pagesize" in the GET request controls the amount of token returned in
            # one request. The "next" entry in "result.value" specifies if there are more token.
            for tfa_token in result['value']['tokens']:
                # disable if hardware token, delete if qr token
                if tfa_token['info']['tokenkind'] == 'hardware':
                    # TODO: refactoring: disable isn't really needed
                    if disable:
                        r = requests.post(url + "/token/disable", json={'serial': tfa_token['serial']}, headers={'Authorization': auth_token})
                        if r.status_code != 200:
                            raise DisableTokenException()
                else:
                    r = requests.delete(url + "/token/%s" % tfa_token['serial'], headers={'Authorization': auth_token})
                    if r.status_code != 200:
                        raise DeleteTokenException()

        # the user was not found. This should rarely / never happen in the frontend, if not for a wrong bind dn/base dn in privacyIDEA
        elif r.status_code == 400 and result['error']['code'] == 904:
            raise UserNotFoundException()
        else:
            raise UnspecifiedErrorException("Error during token delete, http error code {status:s}s\n error message {message:s}".format(
                status=r.status_code,
                message=result['error']['message']
                )
            )
    except ConnectionError:
        raise PrivacyIdeaNotReachableException()


# TODO: only get token if necessary, check token validity.
def get_auth_token(url, username, password):
    try:
        r = requests.post(url + "/auth", json={'username': username, 'password': password})
        if r.status_code != 200:
            raise ValueError("wrong credentials")

        auth_token = r.json()["result"]["value"]["token"]
        if auth_token is None:
            raise ValueError("couldn't extract auth token")

        return auth_token
    except ConnectionError:
        raise PrivacyIdeaNotReachableException()


def rollout_qr_token(url, auth_token, user):
    try:
        if has_active_hardware_token(url, auth_token, user):
            raise QrTokenImpossibleSinceUserHasActiveHardwareTokenException()

        r = requests.post(url + "/token/init", headers={'Authorization': auth_token},
                          json={'genkey': '1', 'user': user, 'type': 'totp', 'hashlib': 'sha1', "pin": ""})
        result = r.json()
        if r.status_code == 400 and result['error']['code'] == 904:
            raise UserNotFoundException()
        elif r.status_code == 200:
            img = result['detail']['googleurl']['img']
            if img is None:
                raise QrTokenGenerationFailedException()

            return {'image': img}
        else:
            raise UnspecifiedErrorException("Unspecified error during token delete %s" % (r.status_code))

    except ConnectionError:
        raise PrivacyIdeaNotReachableException()


def rollover_hardware_token(url, auth_token, user, serial, current_otp):
    try:
        disable_or_delete_all_tokens_for_user(user, url, auth_token)

        # check if token exists
        r = requests.get(url + "/token", headers={'Authorization': auth_token},
                         json={'serial': str(serial)})

        if r.status_code == 404:
            raise TokenNotFoundException()

        res = r.json()['result']['value']

        if res['count'] == 0:
            raise TokenNotFoundException()

        if res['count'] > 1:
            raise TwoTokensWithSameSerialException()

        token = res["tokens"][0]

        # check if token is already assigned to other user
        if token['username'] and token['username'] != user:
            raise TokenAlreadyAssignedToOtherUserException()

        if not otp_valid(url, auth_token, serial, current_otp):
            raise InvalidOtpException()

        # check if token is already assigned to user
        if token['username'] == user:
            # check if token was disabled (might have been lost). Since we check for the current otp
            # there's implicit proof that the token has been recovered and is functional
            if not token['active']:
                r = requests.post(url + "/token/enable", headers={'Authorization': auth_token}, json={'serial': str(serial)})
                result = r.json()

                if r.status_code != 200 or not result or result['value'] is False:
                    raise CouldNotEnableTokenException()
            # Token is already active and assigned to the user
            else:
                raise TokenAlreadyAssignedToUserException()

        else:
            # assign token
            r = requests.post(url + "/token/assign", headers={'Authorization': auth_token},
                              json={'serial': str(serial), 'user': user})

            if r.status_code == 404:
                raise TokenNotFoundException()

            if r.status_code != 200:
                raise TokenCouldNotBeAssignedException()

            result = r.json()
        return result
    except ConnectionError:
        raise PrivacyIdeaNotReachableException()


def lost_token(url, auth_token, user):
    try:
        disable_or_delete_all_tokens_for_user(user, url, auth_token, disable=True, strict=True)
    except ConnectionError:
        raise PrivacyIdeaNotReachableException()


def get_tokens_for_user(url, auth_token, user):
    try:
        r = requests.get(url + "/token", headers={'Authorization': auth_token}, json={'user': user})
        if r.status_code != 200:
            raise ValueError("could not delete token")
        result = r.json()['result']['value']['tokens']
        return result
    except ConnectionError:
        raise PrivacyIdeaNotReachableException()


def get_active_tokens_for_user(url, auth_token, user):
    res = []
    for tfa_token in get_tokens_for_user(url, auth_token, user):
        if tfa_token['active'] is True:
            res.append(tfa_token)

    return res


def has_active_hardware_token(url, auth_token, user):
    res = False

    for tfa_token in get_tokens_for_user(url, auth_token, user):
        if tfa_token['info']['tokenkind'] == 'hardware' and tfa_token['active'] is True:
            res = True

    return res


def otp_valid(url, auth_token, serial, current_otp):
    try:
        r = requests.post(url + "/validate/check", headers={'Authorization': auth_token},
                          json={'serial': str(serial), 'otponly': '1', 'pass': current_otp})
        # 404 -> Token not found, 200 and result['value'] -> wrong otp
        if r.status_code == 404:
            raise TokenNotFoundException()
        elif r.status_code == 200 and r.json()['result']['value'] is True:
            return True
        elif r.status_code == 200 and r.json()['result']['value'] is False:
            return False
    except ConnectionError:
        raise PrivacyIdeaNotReachableException()


def return_hardware_token(url, auth_token, serial, current_otp):
    try:
        if otp_valid(url, auth_token, serial, current_otp):
            # disassociate token
            r = requests.post(url + "/token/unassign", headers={'Authorization': auth_token}, json={'serial': str(serial)})
            if r.status_code != 200:
                raise TokenCouldNotBeUnassignedException()

            # enable token
            r = requests.post(url + "/token/enable", headers={'Authorization': auth_token}, json={'serial': str(serial)})
            if r.status_code != 200:
                raise CouldNotDisableTokenException()

            return "success"
        else:
            raise InvalidOtpException()
    except ConnectionError:
        raise PrivacyIdeaNotReachableException()
